local xyz = {"x","y","z"}
local S = minetest.get_translator("traverse")
local beam_speed = 0.2

local corner_two = vector.new(0.5,0.499,0.5)
local corner_one = vector.new(-0.5,-0.5,-0.5)
local gravity = vector.new(0,-9.81,0)

local has_digilines = minetest.get_modpath("digilines") and true

function remove_mover_if_closer(o,pos) --returns first child of mover or non-mover entity or nil
	local e = o:get_luaentity()
	if e and e.name == "traverse:mover" then
		if (vector.distance(o:get_pos(),pos) < vector.distance(o:get_pos(),e._origin)) then
			local child = o:get_children()[1]
			o:remove()
			return child
		end
	else
		return o
	end
end

minetest.register_entity("traverse:mover", {
	initial_properties = {
		hp_max = 1,
		physical = true,
		collide_with_objects = false,
		collisionbox = {-0.3, -0.3, -0.3, 0.3, 0.3, 0.3},
		textures = {"traverse_transparent.png"},
		static_save = false, --unless properties are serialized (and serializable, attached children aren't,) there is no point in saving the entity
		pointable = false,
		--~ is_visible = false, --buggy, causes child position to not be updated properly

	},
	on_step = function (self,dt,moveresult)
		if moveresult and moveresult.collides then
			self.object:remove()
			return
		end
		
		if self._mode == "beam" then
			local p	 = self.object:get_pos()
			if not vector.in_area(p, self._area[1], self._area[2]) then
				self.object:remove()
				return
			end
			
			self._time = self._time - dt
			if self._time < 0 then
				self._time = self._time + beam_speed
				--check that you're in the traction area and update values from tactor beam like pull and that the beam isn't off
				local node = minetest.get_node(self._origin)
				
				--if beam is off, moved or destroyed, destroy mover
				if node.name ~= "traverse:beam" then
					self.object:remove()
					return
				end
				local meta = minetest.get_meta(self._origin)
				local pull = meta:get_int("pull")==1
				if pull ~= self._pull then
					self.object:remove()
					return
				end
			end
		elseif self._mode == "cannon" then
			local o = self.object
			o:set_acceleration(gravity)
		end
	end,
	--can't serialize attached object, so this is pointless for now
	--~ get_staticdata = function (self)
		--~ return minetest.write_json({
			--~ mode = self._mode,
			--~ origin = self._origin,
			--~ vel = self:get_velocity(),
		--~ })
	--~ end,
	on_activate =  function (self, staticdata, dtime_s)
		if staticdata ~= "" and staticdata ~= nil then
			local data = minetest.parse_json(staticdata) or {}
			self._origin = data.origin
			self._mode = data.mode
		end
		if self._mode == "beam" then
			self._time = beam_speed
			local meta = minetest.get_meta(self._origin)
			self._pull = meta:get_int("pull")==1
			
		end
	end,
	
	on_deactivate = function (self)
		if self._audio then
			minetest.sound_stop(self._audio)
		end
	end,
})

local function is_movable(o)
	if not o:get_attach() then
		local e = o:get_luaentity()
		if o:is_player() or e.name == "__builtin:item" or e.is_mob or e._creatura_mob or e._cmi_is_mob then --mcl, creatura, mobs
			return true
		end
	end
end
local function bool2int(b) return b=="true" and 1 or 0 end
local function int2bool(i) return i==1 and "true" or "false" end
local player_data = {}

local function swap_beam_if_changed(node,pos,to_off)
	local name = node.name
	if name == "traverse:beam" and to_off then 
		node.name="traverse:beam_off"
		minetest.swap_node(pos,node)
	elseif name == "traverse:beam_off" and not to_off then
		node.name="traverse:beam"
		minetest.swap_node(pos,node)
		minetest.get_node_timer(pos):start(0.1)
	end
end

local function adjust_canon(pos,meta,x,y,z)
	if type(x) ~= "number" then return end
	if type(y) ~= "number" then return end
	if type(z) ~= "number" then return end
	meta:set_float("vel_x",x)
	meta:set_float("vel_y",y)
	meta:set_float("vel_z",z)
	local node = minetest.get_node(pos)
	node.param2 = minetest.dir_to_facedir(vector.multiply(vector.new(x,y,z),-1),true)
	minetest.swap_node(pos, node)
end

local on_digiline_receive = function (pos, _, channel, msg)
	local node = minetest.get_node(pos)
	local name = node.name
	local meta = minetest.get_meta(pos)
	if channel == meta:get_string("channel") then
		if type(msg) == "table" then
			if name == "traverse:beam_off" or name == "traverse:beam" then
				if msg.pull ~= nil and type(msg.pull) == "boolean" then
					meta:set_int("pull",msg.pull and 1 or 0)
				end
				if msg.off ~= nil and type(msg.off) == "boolean" then
					swap_beam_if_changed(node,pos,msg.off)
				end
				if msg.reach ~= nil and type(msg.reach) == "number" then
					meta:set_int("reach",msg.reach)
				end
			elseif name == "traverse:cannon" then
				adjust_canon(pos,meta,msg[1],msg[2],msg[3])
			end
		end
	end
end

local digilines =
	{
		receptor = {},
		effector = {
			action = on_digiline_receive
		},
	}

local configure_node = function(pos, node, clicker, itemstack, pointed_thing)
	local node = minetest.get_node(pos)
	local name = node.name
	local meta = minetest.get_meta(pos)
	local player_name = clicker:get_player_name()
	player_data[player_name] = pos
	if name == "traverse:beam" or name == "traverse:beam_off" then
		local str = meta:get_int("reach")
		minetest.show_formspec(player_name, "traverse:form",
		"formspec_version[5]size[4,3]" ..
		"checkbox[0.5,1.2;beam_pull;"..S("Pull")..";"..int2bool(meta:get_int("pull")).."]"..
		"checkbox[0.5,1.6;beam_off;"..S("Off")..";"..(name=="traverse:beam" and "false" or "true").."]"..
		"field[0.5,0.4;3,0.5;beam_reach;"..S("reach")..";"..str.."]"
		..(has_digilines and "field[0.5,2;3,0.5;channel;"..S("Channel")..";"..meta:get_string("channel").."]" or ""))
	elseif name == "traverse:cannon" then
		local x = meta:get_float("vel_x")
		local y = meta:get_float("vel_y")
		local z = meta:get_float("vel_z")
		local str = x..","..y..","..z
		minetest.show_formspec(player_name, "traverse:form",
		"formspec_version[5]size[4,3]" ..
		"field[0.5,0.5;3,0.5;cannon_velocity;"..S("velocity vector")..";"..str.."]"
		..(has_digilines and "field[0.5,2;3,0.5;channel;"..S("Channel")..";"..meta:get_string("channel").."]" or ""))
	end
end

minetest.register_on_player_receive_fields(function(player, formname, fields)
	if formname == "traverse:form" then
		local player_name = player:get_player_name()
		local pos = player_data[player_name]
		local meta = minetest.get_meta(pos)
		if fields.cannon_velocity then
			local xyz = string.split(fields.cannon_velocity,",")
			adjust_canon(pos,meta,tonumber(xyz[1]),tonumber(xyz[2]),tonumber(xyz[3]))
			
		end
		
		if fields.beam_reach then
			if tonumber(fields.beam_reach) then
				meta:set_int("reach",tonumber(fields.beam_reach))
			end
		end
		if fields.beam_pull then
			meta:set_int("pull",bool2int(fields.beam_pull))
		end
		if fields.beam_off then
			local node = minetest.get_node(pos)
			swap_beam_if_changed(node,pos,minetest.is_yes(fields.beam_off))
		end
		if (fields.channel) then
			minetest.get_meta(pos):set_string("channel", fields.channel)
		end
		if fields.quit then
			player_data[player_name] = nil
		end
	end
end)

minetest.register_node("traverse:beam_off", {
	digilines = digilines,
	description = S("Tractor Beam Generator"),
	drop = "traverse:beam",
	tiles = {"default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png^traverse_beam.png"},
	groups = {cracky = 2,not_in_creative_inventory = 1},
	paramtype2 = "facedir",
	on_rightclick = configure_node,
})
minetest.register_node("traverse:beam", {
	digilines = digilines,
	description = S("Tractor Beam Generator"),
	tiles = {"default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png^traverse_beam.png"},
	groups = {cracky = 2},
	paramtype = "light",
	light_source = 12,
	paramtype2 = "facedir",
	on_place = function(itemstack, placer, pointed_thing)
		local player_look_dir = placer:get_look_dir()
		local rotated_facedir = minetest.dir_to_facedir(player_look_dir,true)
		return minetest.item_place(itemstack, placer, pointed_thing,rotated_facedir)
	end,
	on_timer = function (pos)
		local dir = minetest.facedir_to_dir(minetest.get_node(pos).param2)
		--vector like {y = 0, x = -1, z = 0}
		local meta = minetest.get_meta(pos)
		local pull = meta:get_int("pull") == 1
		local reach = meta:get_int("reach")
		
		local forcepos = pos
		if pull then forcepos = vector.add(pos, vector.multiply(dir,-1)) end
		
		
		local reach_search = (reach+0.5)*-1
		local reach_move = (reach+1)*-1
		local objects = minetest.get_objects_in_area(
			{x = forcepos.x + (dir.x<0 and dir.x*reach_search or -0.5), y = forcepos.y + (dir.y<0 and dir.y*reach_search or -0.5)  , z = forcepos.z + (dir.z<0 and dir.z*reach_search or -0.5)},
			{x = forcepos.x + (dir.x>0 and dir.x*reach_search or  0.5), y = forcepos.y + (dir.y>0 and dir.y*reach_search or  0.499), z = forcepos.z + (dir.z>0 and dir.z*reach_search or  0.5)}
		)
		for _, o in pairs(objects) do
			o = remove_mover_if_closer(o,pos)
			
			if o and is_movable(o) then
				local centered = o:get_pos()
				for i=1,3 do
					if dir[i] == 0 then centered[i] = pos[i] end
				end
				
				local staticdata = minetest.write_json({
					mode = "beam",
					origin = pos,
				})
				local mover = minetest.add_entity(centered, "traverse:mover", staticdata)
				
				if o:is_player() then
					mover:get_luaentity()._audio = minetest.sound_play({name = "traverse_beam"},
					{
						to_player = o:get_player_name(),
						loop = true,
					})
				end
				
				o:set_pos(centered)
				o:set_attach(mover)
				local vel = vector.multiply(dir,-5*(pull and -1 or 1))
				--since the default interval limit is 0.2s, 5m/s is the fastest it can go without mobs passing through other beams unnoticed
				mover:set_velocity(vel)
				
				local e = mover:get_luaentity()
				
				e._area = {{x = pos.x + (dir.x<0 and dir.x*reach_move or -0.5), y = pos.y + (dir.y<0 and dir.y*reach_move or -0.5)  , z = pos.z + (dir.z<0 and dir.z*reach_move or -0.5)},
						   {x = pos.x + (dir.x>0 and dir.x*reach_move or  0.5), y = pos.y + (dir.y>0 and dir.y*reach_move or  0.499), z = pos.z + (dir.z>0 and dir.z*reach_move or  0.5)}}
				local area = e._area
				for _,v in ipairs(xyz) do --fix negative areas
					if area[1][v] > area[2][v] then
						local temp = area[1][v]
						area[1][v] = area[2][v]
						area[2][v] = temp
					end
				end
			end
		end
		return true
	end,
	on_construct = function(pos)
		local meta = minetest.get_meta(pos)
		meta:set_int("reach",10)
		minetest.get_node_timer(pos):start(0.1)
	end,
	on_rightclick = configure_node,
})


minetest.register_node("traverse:pipe", {
	description = S("Pipe"),
	tiles = {"traverse_pipe_side.png^[transform2","traverse_pipe_side.png","traverse_pipe_side.png^[transform1","traverse_pipe_side.png^[transform3","traverse_pipe_side.png","traverse_pipe_top.png"},
	drawtype = "nodebox",
	groups = {cracky = 2},
	paramtype2 = "facedir",
	on_place = function(itemstack, placer, pointed_thing)
		local player_look_dir = placer:get_look_dir()
		local rotated_facedir = minetest.dir_to_facedir(player_look_dir,true)
		return minetest.item_place(itemstack, placer, pointed_thing,rotated_facedir)
	end,
	selection_box = {
		type = "fixed",
		fixed = {-0.5,-0.5,-0.5,0.5,0.5,0.5},
	},
	node_box = {
		type = "fixed",
		fixed = {
			{-0.5, -0.5, -5/16,  -5/16,  5/16, -0.5,},
			{5/16, -5/16, -5/16,  0.5, 0.5, -0.5},
			{-5/16, -0.5, -5/16,  0.5, -5/16, -0.5, },
			{-0.5,  5/16, -5/16, 5/16, 0.5, -0.5},
			
			{-7/16, -7/16, -5/16,  -5/16,  5/16, 8/16,},
			{5/16, -5/16, -5/16,  7/16, 7/16, 8/16},
			{-5/16, -7/16, -5/16,  7/16, -5/16, 8/16, },
			{-7/16,  5/16, -5/16, 5/16, 7/16, 8/16},
			
			{-5/16,  -5/16, 0.4, 5/16, 5/16, 0.5},
		},
	},
	collision_box = {
		type = "fixed",
		fixed = {
			{-0.5, -0.5, -0.5,  -7/16,  7/16, 0.5,},
			{7/16, -7/16, -0.5,  0.5, 0.5, 0.5},
			{-7/16, -0.5, -0.5,  0.5, -7/16, 0.5, },
			{-0.5,  7/16, -0.5, 7/16, 0.5, 0.5},
			
			{-7/16,  -7/16, 0.4, 7/16, 7/16, 0.5},
		},
	},
	on_construct = function(pos)
		local function buildable_to(n)
			return minetest.registered_nodes[n.name].buildable_to
		end
		minetest.get_node_timer(pos):start(0.1)
		
		local dir = minetest.facedir_to_dir(minetest.get_node(pos).param2)
		local n
		local v
		local air = 0
		for i=1,100 do
			v = vector.add(pos,vector.multiply(dir,i))
			n = minetest.get_node(v)
			if n.name == "traverse:pipe" then
				break
			end
			if buildable_to(n) then
				air = air + 1
				if air == 3 then
					--clear nodes before pipe if buildable_to
					local before_pipe = vector.add(pos,vector.multiply(dir,-1))
					if buildable_to(minetest.get_node(before_pipe)) then minetest.remove_node(before_pipe) end
					before_pipe = vector.add(pos,vector.multiply(dir,-2))
					if buildable_to(minetest.get_node(before_pipe)) then minetest.remove_node(before_pipe) end
				
					--place second pipe part
					minetest.set_node(
						vector.add(pos,vector.multiply(dir,i-2)),
						{name="traverse:pipe",param2=minetest.dir_to_facedir(vector.multiply(dir, -1),true)}
					)
					--~ --clear buildable_to nodes after pipe
					minetest.remove_node(vector.add(pos,vector.multiply(dir,i-1)))
					minetest.remove_node(vector.add(pos,vector.multiply(dir,i)))
					break
				end
			else
				air = 0
			end
		end
	end,
	after_destruct = function(pos,oldnode)
		local dir = minetest.facedir_to_dir(oldnode.param2)
		local n
		local v
		for i=1,98 do
			v = vector.add(pos,vector.multiply(dir,i))
			n = minetest.get_node(v)
			if n.name == "traverse:pipe" then
				minetest.remove_node(v)
				break
			end
		end
	end,
	on_timer = function (pos)
		local dir = minetest.facedir_to_dir(minetest.get_node(pos).param2)
		--vector like {y = 0, x = -1, z = 0}
		
		local start = vector.add(pos,vector.multiply(dir,-1))
		local objects = minetest.get_objects_in_area(
			vector.add(corner_one,start),
			vector.add(corner_two,start)
		)
		for _, o in pairs(objects) do
			o = remove_mover_if_closer(o,pos)
			if o and is_movable(o) then
				
				local n
				local v
				local air = 0
				for i=1,98 do
					v = vector.add(pos,vector.multiply(dir,i))
					n = minetest.get_node(v)
					if n.name == "traverse:pipe" then
						if minetest.get_node(vector.add(v,vector.multiply(dir,1))).name == "air" and minetest.get_node(vector.add(v,vector.multiply(dir,2))).name == "air" then
							if dir.y == 1 then o:add_velocity({x=4*math.sign(math.random()-0.5),y=10,z=4*math.sign(math.random()-0.5)}) end --prevents player from falling down again when traversing up a pipe
							o:set_pos(vector.add(v,vector.multiply(dir,2)))
							if o:is_player() then
								minetest.sound_play({name = "traverse_pipe",gain=0.5},
								{
									to_player = o:get_player_name(),
								},true)
							end
						end
						break
					else
						air = 0
					end
				end
			end
		end
		return true
	end,
})

minetest.register_node("traverse:cannon", {
	digilines = digilines,
	description = S("Cannon"),
	tiles = {"default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png","default_steel_block.png^traverse_cannon.png"},
	drawtype = "nodebox",
	selection_box = {
		type = "fixed",
		fixed = {-6/16, -6/16, -0.5, 6/16, 6/16, 0.5},
	},
	node_box = {
		type = "fixed",
		fixed = {
			{-4/16, -4/16, -0.5, 4/16, 4/16, 0.5},
			{-6/16, -6/16, -3/16, 6/16, 6/16, 6/16}
		},
	},
	walkable = false, --required or player will sometimes not leave the cannon when shot
	groups = {cracky = 2},
	paramtype2 = "facedir",
	on_place = function(itemstack, placer, pointed_thing)
		local player_look_dir = placer:get_look_dir()
		local rotated_facedir = minetest.dir_to_facedir(player_look_dir,true)
		return minetest.item_place(itemstack, placer, pointed_thing,rotated_facedir)
	end,

	on_timer = function (pos)
		local block_pos=pos
		local objects = minetest.get_objects_in_area(
			{x = block_pos.x - 1, y = block_pos.y - 1, z = block_pos.z - 1},
			{x = block_pos.x + 1, y = block_pos.y + 1, z = block_pos.z + 1}
		)
		for _, o in pairs(objects) do
			o = remove_mover_if_closer(o,pos)
			
			if o and is_movable(o) then
				
				local staticdata = minetest.write_json({
					mode = "cannon",
					origin = pos,
				})
				local mover = minetest.add_entity(pos, "traverse:mover", staticdata)
				mover:set_pos(pos)
				o:set_pos(pos)
				o:set_attach(mover)
				
				local meta = minetest.get_meta(pos)
				local x = meta:get_float("vel_x")
				local y = meta:get_float("vel_y")
				local z = meta:get_float("vel_z")
				
				mover:add_velocity(vector.new(x,y,z))

				
				minetest.sound_play({name = "traverse_cannon"},
				{
					pos = o:get_pos(),
				})
			end
		end
		return true
	end,
	on_construct = function(pos)
		local v = vector.multiply(
				minetest.facedir_to_dir(minetest.get_node(pos).param2)
,				-10
			)
		if v.y == 0 then
			v.y = 5
		else
			v.x = 2
		end
		local meta = minetest.get_meta(pos)
		meta:set_float("vel_x",math.round(v.x))
		meta:set_float("vel_y",math.round(v.y))
		meta:set_float("vel_z",math.round(v.z))
		
		minetest.get_node_timer(pos):start(0.1)
	end,
	on_rightclick = configure_node,
})